var async = require('async');
var assert = require('assert');
var http = require('http');
var dgram = require('dgram');
var querystring = require('querystring');

var defaults = require('../broker').defaults;
var MetricsBroker = require('../broker').MetricsBroker;
var client = require('../broker').client;

var Counter = require('../instruments').Counter;
var Timer = require('../instruments').Timer;
var Histogram = require('../instruments').Histogram;

var broker;


// Set to true to exhaustively test broker (will cause broker errors to be
// emitted to stderr)
var verbose = false;


/**
 * Invoke the specified callback after each of the specified events has been
 * invoked on the specified object.
 */
function waitForEvents(obj, events, callback) {
  var e,
      eventTable = {};

  while (e = events.pop()) {
    eventTable[e] = true;
    (function(e) {
      obj.once(e, function() {
        delete eventTable[e];
        if (Object.keys(eventTable).length === 0) {
          callback();
        }
      });
    })(e);
  }
}


function getMetrics(callback) {
  var body = '';
  http.get({ host: 'localhost',
             port: defaults.DEFAULT_SERVER_PORT,
             path: '/metrics' },
             function(res) {
               assert.strictEqual(res.statusCode, 200, 'HTTP status code should be 200');
               res.setEncoding('utf8');
               res.on('data', function(chunk) {  body = body + chunk; });
               res.on('end', function() { callback(null, JSON.parse(body)); });
           });
}


exports.setUp = function(callback) {
  broker = new MetricsBroker();
  waitForEvents(broker, ['brokerReady', 'serverReady'], callback);
};


exports.testGarbage = function(callback) {
  if (verbose) {
    var client = dgram.createSocket('unix_dgram');
    async.waterfall([
      function(callback) {
        var msg = new Buffer('bad');
        client.send(msg, 0, msg.length, defaults.DEFAULT_BROKER_SOCK, callback);
      },
      function(sentBytes, callback) {
        var msg = new Buffer('{}');
        client.send(msg, 0, msg.length, defaults.DEFAULT_BROKER_SOCK, callback);
      },
      function(sentBytes, callback) {
        var msg = new Buffer('{"foo":17}');
        client.send(msg, 0, msg.length, defaults.DEFAULT_BROKER_SOCK, callback);
      },
    ], function(err) {
      assert.ifError(err);
      callback();
    });
  } else {
    callback();
  }
};


exports.testCounter = function(callback) {
  var dimensions = { cust: 'custA', host: 'hostQ' };
  var key = querystring.stringify(dimensions);
  var name = 'testCounter';
  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) { Counter.inc(name, dimensions, 1, callback); },
    function(callback) { Counter.inc(name, dimensions, 2, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 1, 'Data must have one dimension');
      assert.ok(data.hasOwnProperty(key), 'Received data does not contain proper dimensions');
      var metric = data[key][name];
      assert.strictEqual(metric.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metric.count, 3, 'Received data has wrong count');
      callback();
    },
    function(callback) { Counter.clear(name, dimensions, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      var metric = data[key][name];
      assert.strictEqual(metric.count, 0, 'Received data has wrong count');
      callback();
    },
    // test wraparound
    function(callback) { Counter.inc(name, dimensions, Math.pow(2,32), callback); },
    function(callback) { Counter.inc(name, dimensions, 2, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      var metric = data[key][name];
      assert.strictEqual(metric.count, 1, 'Received data has wrong count');
      callback();
    },
    function(callback) { Counter.clear(name, dimensions, callback); },
    // test dec below 0
    function(callback) { Counter.dec(name, dimensions, 1, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      var metric = data[key][name];
      assert.strictEqual(metric.count, 0, 'Received data has wrong count');
      callback();
    },
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};

exports.testUTF8 = function(callback) {
  var dimensions = { '\u00C1\u0072\u0076\u00ED\u007A\u0074\u0171\u0072\u0151': '\u3044\u308D\u306F\u306B\u307B\u3078\u3069',
    '\u9154\u3072\u3082\u305B\u305A': '\u0447\u0430\u044E' };
  var key = querystring.stringify(dimensions);
  var name = '\u0050\u0063\u0068\u006E\u0105\u0107';
  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) { Counter.inc(name, dimensions, 1, callback); },
    function(callback) { Counter.inc(name, dimensions, 2, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 1, 'Data must have one dimension');
      assert.ok(data.hasOwnProperty(key), 'Received data does not contain proper dimensions');
      var metric = data[key][name];
      assert.strictEqual(metric.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metric.count, 3, 'Received data has wrong count');
      callback();
    },
    function(callback) { Counter.clear(name, dimensions, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      var metric = data[key][name];
      assert.strictEqual(metric.count, 0, 'Received data has wrong count');
      callback();
    },
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};

exports.testTimer = function(callback) {
  var dimensions = { cust: 'custA', host: 'hostQ' };
  var key = querystring.stringify(dimensions);
  var name = 'testTimer';
  var timer = new Timer(name, dimensions);
  var count = 10;
  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) {
      var i = 0;
      async.whilst(
        function() { return i < 10; },
        function(callback) {
          timer.start(); 
          setTimeout(function() { timer.stop(); i++; callback(); }, 100);
        },
        function(err) { callback(err); }
      );
    },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 1, 'Data must have one dimension');
      assert.ok(data.hasOwnProperty(key), 'Received data does not contain proper dimensions');
      var metric = data[key][name];
      assert.strictEqual(metric.type, 'histogram', 'Incorrect data type');
      assert.strictEqual(metric.count, count, 'Incorrect count');
      assert.ok(metric.sum < 2, 'Duration sum much too high');
      assert.ok(metric.min >= 0.1 , 'Duration min too low');
      // Requires bugfixes to metrics
      assert.ok(metric.mean <= 0.2, 'Duration mean too high');
      assert.ok(metric.median <= 0.2, 'Duration median too high');
      callback();
    }
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};


exports.testHistogram = function(callback) {
  var dimensions = { cust: 'custA', host: 'hostQ' };
  var key = querystring.stringify(dimensions);
  var name = 'testHistogram';
  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) { Histogram.update(name, dimensions, 1, callback); },
    function(callback) { Histogram.update(name, dimensions, 50, callback); },
    function(callback) { Histogram.update(name, dimensions, 98, callback); },
    function(callback) { Histogram.update(name, dimensions, 99, callback); },
    function(callback) { Histogram.update(name, dimensions, 100, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 1, 'Data must have one dimension');
      assert.ok(data.hasOwnProperty(key), 'Received data does not contain proper dimensions');
      var metric = data[key][name];
      assert.strictEqual(metric.type, 'histogram', 'Incorrect data type');
      assert.strictEqual(metric.count, 5, 'Incorrect count');
      assert.strictEqual(metric.sum, 348, 'Incorrect sum');
      // Requires bugfixes to metrics
      // assert.strictEqual(metric.mean, 69.6, 'Incorrect mean');
      // assert.strictEqual(metric.median, 98, 'Incorrect median');
      callback();
    },
    function(callback) { Histogram.clear(name, dimensions, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      var metric = data[key][name];
      // Requires bugfixes to metrics
      // assert.strictEqual(metric.count, 0, 'Received data has wrong count');
      callback();
    },
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};


exports.testMultiInstruments = function(callback) {
  var dimensions = { cust: 'custA', host: 'hostQ' };
  var key = querystring.stringify(dimensions);
  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) { Counter.inc('counterA', dimensions, 1, callback); },
    function(callback) { Counter.inc('counterB', dimensions, 1, callback); },
    function(callback) { Counter.inc('counterA', dimensions, 1, callback); },
    function(callback) { Counter.inc('counterB', dimensions, 1, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 1, 'Data must have one dimension');
      assert.ok(data.hasOwnProperty(key), 'Received data does not contain proper dimensions');
      var metrics = data[key];
      assert.ok(metrics.hasOwnProperty('counterA'), 'Received data does not contain counterA');
      assert.ok(metrics.hasOwnProperty('counterB'), 'Received data does not contain counterB');
      assert.strictEqual(metrics.counterA.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metrics.counterB.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metrics.counterA.count, 2, 'Incorrect value for counterA');
      assert.strictEqual(metrics.counterB.count, 2, 'Incorrect value for counterB');
      callback();
    }
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};


exports.testEmptyDimensions = function(callback) {
  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) { Counter.inc('counterA', null, 1, callback); },
    function(callback) { Histogram.update('histA', null, 1, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 1, 'Data must have one dimension');
      assert.ok(data.hasOwnProperty(''), 'Received data does not contain proper dimensions');
      assert.strictEqual(data[''].counterA.type, 'counter', 'Incorrect data type');
      callback();
    }
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};


exports.testMultiDimensions = function(callback) {
  var dimA = { cust: 'custA', host: 'hostQ' };
  var keyA = querystring.stringify(dimA);
  var dimB = { cust: 'custB', host: 'hostZ' };
  var keyB = querystring.stringify(dimB);
  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) { Counter.inc('counterA', dimA, 1, callback); },
    function(callback) { Counter.inc('counterB', dimA, 1, callback); },
    function(callback) { Counter.inc('counterA', dimB, 2, callback); },
    function(callback) { Counter.inc('counterB', dimB, 2, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 2, 'Data must have 2 dimensions');
      assert.ok(data.hasOwnProperty(keyA), 'Received data does not contain proper dimensions');
      assert.ok(data.hasOwnProperty(keyB), 'Received data does not contain proper dimensions');
      var metricA = data[keyA];
      var metricB = data[keyB];
      assert.strictEqual(metricA.counterA.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metricA.counterB.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metricB.counterA.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metricB.counterB.type, 'counter', 'Incorrect data type');
      assert.strictEqual(metricA.counterA.count, 1, 'Incorrect value for counterA');
      assert.strictEqual(metricA.counterB.count, 1, 'Incorrect value for counterB');
      assert.strictEqual(metricB.counterA.count, 2, 'Incorrect value for counterA');
      assert.strictEqual(metricB.counterB.count, 2, 'Incorrect value for counterB');
      callback();
    }
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};


exports.testDuplicateNames = function(callback) {
  if (verbose) {
    // This test causes noise to be emitted to stderr, so skip it to avoid
    // confusing users.  Set to true to activate the test.
      var dimensions = { cust: 'custA', host: 'hostQ' };
      var name = 'uniqueName';
      async.waterfall([
        function(callback) { client.clear(callback); },
        function(callback) { Counter.inc(name, dimensions, 1, callback); },
        function(callback) { Histogram.update(name, dimensions, 1, callback); }
      ], function(err) {
        assert.ifError(err);
        callback();
      });
  } else {
    callback();
  }
};


exports.testDimensionOrder = function(callback) {
  var name = 'testCounter',
      cust = 'custA',
      host = 'hostQ';

  async.waterfall([
    function(callback) { client.clear(callback); },
    function(callback) { Counter.inc(name, { cust: cust, host: host}, 1, callback); },
    function(callback) { Counter.inc(name, { host: host, cust: cust}, 1, callback); },
    function(callback) { getMetrics(callback); },
    function(data, callback) {
      assert.strictEqual(Object.keys(data).length, 1, 'Data must have 1 dimension');
      assert.ok(data.hasOwnProperty('cust=custA&host=hostQ'), 'Received data does not contain proper dimensions');
      assert.strictEqual(data['cust=custA&host=hostQ'].testCounter.count, 2, 'Received data has wrong count');
      callback();
    },
  ], function(err) {
    assert.ifError(err);
    callback();
  });
};


exports.tearDown = function(callback) {
  broker.shutDown();
  waitForEvents(broker, ['serverClose', 'brokerClose'], callback);
};


function runAllTests() {
  var test,
      tests = [];
  for (test in exports) {
    if (exports.hasOwnProperty(test) &&
        test != 'setUp' &&
        test != 'tearDown') {
      tests.push(exports[test]);
    }
  }
  if (exports.hasOwnProperty('setUp')) {
    tests.unshift(exports.setUp);
  }
  if (exports.hasOwnProperty('tearDown')) {
    tests.push(exports.tearDown);
  }
  async.series(tests);
}

if (require.main === module) {
  runAllTests();
}
